"""ExpandObjects module"""
import logging as lg
import shutil
import subprocess
import time
from io import StringIO
from subprocess import CalledProcessError
from threading import Thread

from packaging.version import Version
from path import Path
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from archetypal.eplus_interface.energy_plus import EnergyPlusProgram
from archetypal.utils import log


class ExpandObjectsExe(EnergyPlusProgram):
    """ExpandObject Wrapper"""

    def __init__(self, idf):
        """Constructor."""
        super().__init__(idf)

    @property
    def cmd(self):
        """Get the command."""
        return ["ExpandObjects"]


class ExpandObjectsThread(Thread):
    """ExpandObjects program manager."""

    def __init__(self, idf, tmp):
        """Constructor."""
        super(ExpandObjectsThread, self).__init__()
        self.p = None
        self.std_out = None
        self.std_err = None
        self.idf = idf
        self.cancelled = False
        self.run_dir = Path("")
        self.exception = None
        self.name = "ExpandObjects_" + self.idf.name
        self.tmp = tmp
        self.cmd = None

    def run(self):
        """Wrapper around the EnergyPlus command line interface."""
        try:
            self.cancelled = False
            # get version from IDF object or by parsing the IDF file for it

            # Move files into place
            tmp = self.tmp
            self.epw = self.idf.epw.copy(tmp / "in.epw").expand()
            self.idfname = Path(self.idf.savecopy(tmp / "in.idf")).expand()
            self.idd = self.idf.iddname.copy(tmp / "Energy+.idd").expand()
            expand_object_exe = shutil.which("ExpandObjects", path=self.eplus_home)
            self.expandobjectsexe = Path(expand_object_exe).copy2(tmp)
            self.run_dir = Path(tmp).expand()

            # Run ExpandObjects Program
            with logging_redirect_tqdm(loggers=[lg.getLogger(self.idf.name)]):
                self.cmd = ExpandObjectsExe(self.idf).cmd
                with tqdm(
                    unit_scale=True,
                    miniters=1,
                    desc=f"{expand_object_exe} #{self.idf.position}-{self.idf.name}",
                    position=self.idf.position,
                ) as progress:

                    self.p = subprocess.Popen(
                        self.cmd,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        shell=True,  # can use shell
                        cwd=self.run_dir.abspath(),
                    )
                    start_time = time.time()
                    # self.msg_callback("ExpandObjects started")
                    for line in self.p.stdout:
                        self.msg_callback(line.decode("utf-8").strip("\n"))
                        progress.update()

                    # We explicitly close stdout
                    self.p.stdout.close()

                    # Wait for process to complete
                    self.p.wait()

                    # Communicate callbacks
                    if self.cancelled:
                        self.msg_callback("ExpandObjects cancelled")
                        # self.cancelled_callback(self.std_out, self.std_err)
                    else:
                        if self.p.returncode == 0:
                            self.msg_callback(
                                f"ExpandObjects completed in "
                                f"{time.time() - start_time:,.2f} seconds"
                            )
                            self.success_callback()
                        else:
                            self.failure_callback()
        except Exception as e:
            self.exception = e
            if self.p is not None:
                self.p.kill()  # kill process to be sure
            return

    def msg_callback(self, *args, **kwargs):
        """Pass message to logger."""
        log(*args, name=self.idf.name, **kwargs)

    def success_callback(self):
        """Replace idf with expanded.idf."""
        if (self.run_dir / "expanded.idf").exists():
            if isinstance(self.idf.idfname, StringIO):
                file = StringIO((self.run_dir / "expanded.idf").read_text())
            else:
                file = (self.run_dir / "expanded.idf").copy(self.idf.idfname)
            self.idf.idfname = file
        if (Path(self.run_dir) / "GHTIn.idf").exists():
            self.idf.include.append(
                (Path(self.run_dir) / "GHTIn.idf").copy(
                    self.idf.output_directory.makedirs_p() / "GHTIn.idf"
                )
            )

        if (Path(self.run_dir) / "BasementGHTIn.idf").exists():
            self.idf.include.append(
                (Path(self.run_dir) / "BasementGHTIn.idf").copy(
                    self.idf.output_directory.makedirs_p() / "BasementGHTIn.idf"
                )
            )

    def failure_callback(self):
        """Read stderr and pass to logger."""
        for line in self.p.stderr:
            self.msg_callback(line.decode("utf-8"), level=lg.ERROR)
        self.exception = CalledProcessError(
            self.p.returncode, cmd=self.cmd, stderr=self.p.stderr
        )

    def cancelled_callback(self, stdin, stdout):
        """Call on cancelled."""
        pass

    @property
    def eplus_home(self):
        """Get the version-dependant directory where executables are installed."""
        if self.idf.file_version <= Version("7.2"):
            install_dir = self.idf.file_version.current_install_dir / "bin"
        else:
            install_dir = self.idf.file_version.current_install_dir
        return install_dir

    def stop(self):
        if self.p.poll() is None:
            self.msg_callback("Attempting to cancel simulation ...")
            self.cancelled = True
            self.p.kill()



import logging as lg
import subprocess
import time
from subprocess import CalledProcessError
from threading import Thread

import eppy
from eppy.runner.run_functions import paths_from_version
from path import Path
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from archetypal.eplus_interface.exceptions import (
    EnergyPlusProcessError,
    EnergyPlusVersionError,
)
from archetypal.eplus_interface.version import EnergyPlusVersion
from archetypal.utils import log


class EnergyPlusProgram:
    def __init__(self, idf):
        self.idf = idf

    @property
    def eplus_home(self):
        eplus_exe, eplus_home = paths_from_version(self.idf.as_version.dash)
        if not Path(eplus_home).exists():
            raise EnergyPlusVersionError(
                msg=f"No EnergyPlus Executable found for version "
                f"{EnergyPlusVersion(self.idf.as_version)}"
            )
        else:
            return Path(eplus_home)


class EnergyPlusExe:
    """Usage: energyplus [options] [input-file]"""

    def __init__(
        self,
        idfname,
        epw,
        output_directory,
        ep_version,
        annual=False,
        convert=False,
        design_day=False,
        help=False,
        idd=None,
        epmacro=False,
        output_prefix="eplus",
        readvars=True,
        output_sufix="L",
        version=False,
        expandobjects=True,
    ):
        """
        Args:
            idfname (str):
            epw (str): Weather file path (default: in.epw in current directory))
            output-directory (str): Output directory path (default: current directory)
            ep_version (archetypal.EnergyPlusVersion): The version of energyplus
                executable.
            annual (bool): Force annual simulation. (default: False)
            convert (bool): Output IDF->epJSON or epJSON->IDF, dependent on input
                file type. (default: False)
            design_day (bool): Force design-day-only simulation. (default: False)
            help (bool): Display help information
            idd (str) :Input data dictionary path (default: Energy+.idd in
                executable directory)
            epmacro (bool): Run EPMacro prior to simulation. (default: True)
            output_prefix (str): Prefix for output file names (default: eplus)
            readvars (bool): Run ReadVarsESO after simulation. (default: True)
            output_suffix (str): Suffix style for output file names (
                default: L)
                    -L: Legacy (e.g., eplustbl.csv)
                    -C: Capital (e.g., eplusTable.csv)
                    -D: Dash (e.g., eplus-table.csv)
            version (bool): Display version information (default: False)
            expandobjects (bool): Run ExpandObjects prior to simulation. (default: True)
        """
        self.a = annual
        self.c = convert
        self.d = output_directory
        self.D = design_day
        self.h = help
        self.i = idd
        self.m = epmacro
        self.p = output_prefix
        self.r = readvars
        self.s = output_sufix
        self.v = version
        self.w = epw
        self.x = expandobjects

        self.idfname = idfname
        self.ep_version = ep_version

        self.get_exe_path()

    def get_exe_path(self):
        (eplus_exe_path, eplus_weather_path) = eppy.runner.run_functions.install_paths(
            self.ep_version.dash, self.i
        )
        if not Path(eplus_exe_path).exists():
            raise EnergyPlusVersionError(
                msg=f"No EnergyPlus Executable found for version "
                f"{EnergyPlusVersion(self.ep_version)}"
            )
        self.eplus_exe_path = Path(eplus_exe_path).expand()
        self.eplus_weather_path = Path(eplus_weather_path).expand()

    def __str__(self):
        """Return string representation."""
        return " ".join(self.__repr__())

    def __repr__(self):
        """Return a representation of self."""
        cmd = [self.eplus_exe_path]
        for key, value in self.__dict__.items():
            if key not in [
                "idfname",
                "ep_version",
                "eplus_exe_path",
                "eplus_weather_path",
            ]:
                if isinstance(value, bool):
                    cmd.append(f"-{key}") if value else None
                else:
                    cmd.extend([f"-{key}", value]) if value is not None else None
        cmd.append(self.idfname)
        return cmd

    def cmd(self):
        return self.__repr__()


class EnergyPlusThread(Thread):
    def __init__(self, idf, tmp):
        """

        Args:
            idf (IDF): The idf model.
            tmp (str or Path): The directory in which the process will be launched.
        """
        super(EnergyPlusThread, self).__init__()
        self.p = None
        self.std_out = None
        self.std_err = None
        self.idf = idf
        self.cancelled = False
        self.run_dir = Path("")
        self.exception = None
        self.name = "EnergyPlus_" + self.idf.name
        self.tmp = tmp

    def stop(self):
        if self.p.poll() is None:
            self.msg_callback("Attempting to cancel simulation ...")
            self.cancelled = True
            self.p.kill()

    def run(self):
        """Wrapper around the EnergyPlus command line interface.

        Adapted from :func:`eppy.runner.runfunctions.run`.
        """
        self.cancelled = False
        # get version from IDF object or by parsing the IDF file for it

        tmp = self.tmp
        self.epw = Path(self.idf.epw.copy(tmp)).expand()
        self.idfname = Path(self.idf.savecopy(tmp / self.idf.name)).expand()
        self.idd = Path(self.idf.iddname.copy(tmp)).expand()
        self.run_dir = Path(tmp).expand()
        self.include = [Path(file).copy(tmp) for file in self.idf.include]

        # build a list of command line arguments
        try:
            eplus_exe = EnergyPlusExe(
                idfname=self.idfname,
                epw=self.epw,
                output_directory=self.run_dir,
                ep_version=self.idf.file_version,
                annual=self.idf.annual,
                convert=self.idf.convert,
                design_day=self.idf.design_day,
                help=False,
                idd=self.idd,
                epmacro=self.idf.epmacro,
                output_prefix=self.idf.output_prefix,
                readvars=self.idf.readvars,
                output_sufix=self.idf.output_suffix,
                version=False,
                expandobjects=self.idf.expandobjects,
            )
            self.cmd = eplus_exe.cmd()
        except EnergyPlusVersionError as e:
            self.exception = e
            self.p.kill()  # kill process to be sure
            return
        with logging_redirect_tqdm(loggers=[lg.getLogger(self.idf.name)]):
            # Start process with tqdm bar
            with tqdm(
                unit_scale=False,
                total=self.idf.energyplus_its if self.idf.energyplus_its > 0 else None,
                miniters=1,
                desc=f"{eplus_exe.eplus_exe_path} #{self.idf.position}-{self.idf.name}"
                if self.idf.position
                else f"{eplus_exe.eplus_exe_path} {self.idf.name}",
                position=self.idf.position,
            ) as progress:
                self.p = subprocess.Popen(
                    self.cmd,
                    shell=False,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                start_time = time.time()
                self.msg_callback("Simulation started")
                self.idf._energyplus_its = 0  # reset counter
                for line in self.p.stdout:
                    self.msg_callback(line.decode("utf-8").strip("\n"))
                    self.idf._energyplus_its += 1
                    progress.update()

                # We explicitly close stdout
                self.p.stdout.close()

                # Wait for process to complete
                self.p.wait()

                # Communicate callbacks
                if self.cancelled:
                    self.msg_callback("Simulation cancelled")
                    self.cancelled_callback(self.std_out, self.std_err)
                else:
                    if self.p.returncode == 0:
                        self.msg_callback(
                            "EnergyPlus Completed in {:,.2f} seconds".format(
                                time.time() - start_time
                            )
                        )
                        self.success_callback()
                    else:
                        self.msg_callback("Simulation failed")
                        self.failure_callback()

    def msg_callback(self, *args, **kwargs):
        msg, *_ = args
        for m in msg.split("\r"):
            if m:
                log(m, name=self.idf.name, **kwargs)

    def success_callback(self):
        save_dir = self.idf.simulation_dir
        if self.idf.keep_data:
            try:
                save_dir.rmtree_p()  # purge target dir
                self.run_dir.copytree(save_dir)  # copy files
            except PermissionError as e:
                pass
            else:
                log(
                    "Files generated at the end of the simulation: %s"
                    % "\n".join(save_dir.files()),
                    lg.DEBUG,
                    name=self.name,
                )

    def failure_callback(self):
        error_filename = self.run_dir / self.idf.output_prefix + "out.err"
        try:
            with open(error_filename, "r") as stderr:
                stderr_r = stderr.read()
            if self.idf.keep_data_err:
                failed_dir = self.idf.simulation_dir.mkdir_p()
                try:
                    failed_dir.rmtree_p()
                except PermissionError as e:
                    log(f"Could not remove {failed_dir}")
                else:
                    self.run_dir.copytree(failed_dir)  # no need to create folder before
            self.exception = EnergyPlusProcessError(
                cmd=self.cmd, stderr=stderr_r, idf=self.idf
            )
        except FileNotFoundError:
            self.exception = CalledProcessError(
                self.p.returncode, cmd=self.cmd, stderr=self.p.stderr
            )

    def cancelled_callback(self, stdin, stdout):
        pass

    @property
    def eplus_home(self):
        eplus_exe, eplus_home = paths_from_version(self.idf.as_version.dash)
        if not Path(eplus_home).exists():
            raise EnergyPlusVersionError(
                msg=f"No EnergyPlus Executable found for version "
                f"{EnergyPlusVersion(self.idf.as_version)}"
            )
        else:
            return Path(eplus_home)



import numpy as np
import pandas as pd
import pytest

import archetypal.settings as settings
from archetypal import IDF
from archetypal.eplus_interface import EnergyPlusVersion
from archetypal.schedule import Schedule, ScheduleTypeLimits
from archetypal.template.schedule import UmiSchedule, YearSchedule
from archetypal.utils import config


class TestScheduleTypeLimits:
    """Test ScheduleTypeLimits class."""

    def test_from_to_dict(self):
        data = {
            "Name": "Fractional",
            "LowerLimit": 0,
            "UpperLimit": 1,
            "NumericType": None,
            "UnitType": "Dimensionless",
        }

        type_limit = ScheduleTypeLimits.from_dict(data)
        type_limit_dict = type_limit.to_dict()
        type_limit_dup = ScheduleTypeLimits.from_dict(type_limit_dict)

        assert type_limit.Name == type_limit_dup.Name == "Fractional"

    def test_from_epbunch(self):
        idf = IDF()
        epbunch = idf.anidfobject("SCHEDULETYPELIMITS", Name="Fractional")

        type_limit = ScheduleTypeLimits.from_epbunch(epbunch)
        assert type_limit


class TestSchedule:
    @pytest.fixture()
    def schedules_in_necb_specific(self, config):
        idf = IDF(
            "tests/input_data/necb/NECB 2011-MediumOffice-NECB HDD "
            "Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf"
        )
        epbunch = idf.schedules_dict["NECB-A-Thermostat Setpoint-Heating".upper()]
        s = Schedule.from_epbunch(epbunch, start_day_of_the_week=0)
        yield s

    def test_scale(self, schedules_in_necb_specific):
        before_sum = sum(schedules_in_necb_specific.Values)
        assert pytest.approx(
            before_sum, sum(schedules_in_necb_specific.scale(0.1).Values)
        )

    def test_plot(self, schedules_in_necb_specific):
        schedules_in_necb_specific.plot(drawstyle="steps-post")

    def test_plot2d(self, schedules_in_necb_specific):
        schedules_in_necb_specific.plot2d(show=False, save=False)

    def test_make_umi_schedule(self):
        """Test only a single schedule name."""

        idf = IDF("tests/input_data/schedules/schedules.idf", prep_outputs=False)
        ep_bunch = idf.schedules_dict["CoolingCoilAvailSched".upper()]
        s = UmiSchedule.from_epbunch(ep_bunch, start_day_of_the_week=0)
        new = s.develop()
        assert hash(s) != hash(new)
        assert id(s) != id(new)

        assert isinstance(s, UmiSchedule)
        assert isinstance(new, YearSchedule)
        assert len(s.all_values) == len(new.all_values)
        np.testing.assert_array_equal(new.all_values, s.all_values)

    def test_constant_schedule(self):
        const = Schedule.constant_schedule()
        assert const.__class__.__name__ == "Schedule"

    @pytest.fixture()
    def new_idf(self, config):
        yield IDF(prep_outputs=False)

    def test_from_values(self, new_idf):
        import numpy as np

        heating_sched = UmiSchedule.from_values(
            Name="Zone_Heating_Schedule",
            Values=np.ones(8760),
            Type="Fraction",
            idf=idf,
        )
        assert len(heating_sched.all_values) == 8760

    def test_replace(self):
        """Test replacing values while keeping full load hours constant."""
        sch = Schedule.from_values("Test", [1] * 6 + [0.5] * 12 + [1] * 6)
        new = pd.Series([1, 1], index=sch.series.index[11:13])

        orig = sch.series.sum()

        sch.replace(new)

        new = sch.series.sum()

        # assert the full load hours (sum) has not changed.
        assert new == pytest.approx(orig)


idf_file = "tests/input_data/schedules/test_multizone_EP.idf"


def schedules_idf():
    config(cache_folder="tests/.temp/cache")
    idf = IDF(
        idf_file,
        epw="tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw",
        readvars=True,
        include=[
            EnergyPlusVersion.current().current_install_dir
            / "DataSets"
            / "TDV"
            / "TDV_2008_kBtu_CTZ06.csv"
        ],
    )
    return idf


idf = schedules_idf()
schedules_dict = idf._get_all_schedules(yearly_only=True)
schedules = list(schedules_dict.values())
ids = [i.replace(" ", "_") for i in schedules_dict.keys()]


@pytest.fixture(scope="module")
def csv_out(config):
    idf = schedules_idf().simulate()
    csv, *_ = idf.simulation_dir.files("*out.csv")
    yield csv


schedules = [
    pytest.param(
        schedule,
        marks=pytest.mark.xfail(
            reason="Can't quite capture all possibilities with special days"
        ),
    )
    if schedule == "POFF"
    else pytest.param(schedule, marks=pytest.mark.xfail(raises=NotImplementedError))
    if schedule == "Cooling Setpoint Schedule"
    else schedule
    for schedule in schedules
]


@pytest.fixture(params=schedules, ids=ids, scope="module")
def schedule_parametrized(request, csv_out):
    """Create the test_data"""
    import pandas as pd

    ep_bunch = request.param
    origin = Schedule.from_epbunch(ep_bunch)

    # create year:week:day version
    new_eps = origin.to_year_week_day()
    new = origin

    index = origin.series.index
    epv = pd.read_csv(csv_out)
    epv.columns = epv.columns.str.strip()
    epv = epv.loc[:, ep_bunch.Name.upper() + ":Schedule Value [](Hourly)"].values
    expected = pd.Series(epv, index=index)

    print("Year: {}".format(new_eps[0].Name))
    print("Weeks: {}".format([obj.Name for obj in new_eps[1]]))
    print("Days: {}".format([obj.Name for obj in new_eps[2]]))

    yield origin, new, expected


def test_ep_versus_schedule(schedule_parametrized):
    """Main test. Will run the idf using EnergyPlus, retrieve the csv file,
    create the schedules and compare"""

    orig, new, expected = schedule_parametrized

    # slice_ = ('2018/01/01 00:00', '2018/01/08 00:00')  # first week
    # slice_ = ('2018/05/20 12:00', '2018/05/22 12:00')
    slice_ = ("2018-10-05 12:00", "2018-10-07 12:00")  # Holiday
    # slice_ = ('2018/01/01 00:00', '2018/12/31 23:00')  # all year
    # slice_ = ('2018/04/30 12:00', '2018/05/01 12:00')  # break

    mask = expected.values.round(3) != orig.all_values.round(3)

    # # region Plot
    # fig, ax = plt.subplots(1, 1, figsize=(5, 4))
    # orig.plot(slice=slice_, ax=ax, legend=True, drawstyle='steps-post',
    #           linestyle='dashed')
    # new.plot(slice=slice_, ax=ax, legend=True, drawstyle='steps-post',
    #          linestyle='dotted')
    # expected.loc[slice_[0]:slice_[1]].plot(label='E+', legend=True, ax=ax,
    #                                        drawstyle='steps-post',
    #                                        linestyle='dashdot')
    # ax.set_title(orig.Name.capitalize())
    # plt.show()
    # # endregion

    print(pd.DataFrame({"actual": orig.series[mask], "expected": expected[mask]}))
    np.testing.assert_array_almost_equal(orig.all_values, expected, verbose=True)
    np.testing.assert_array_almost_equal(new.all_values, expected, verbose=True)



"""archetypal GasLayer."""
import collections
import logging as lg
import math

from sigfig import round
from validator_collection import validators

from archetypal.utils import log


class GasLayer(object):
    """Class used to define one gas layer in a window construction assembly.

    This class has two attributes:

    1. Material (:class:`GasMaterial`): the material object for this layer.
    2. Thickness (float): The thickness of the material in the layer.
    """

    __slots__ = ("_material", "_thickness")

    def __init__(self, Material, Thickness, **kwargs):
        """Initialize a MaterialLayer object with parameters.

        Args:
            Material (GasMaterial):
            Thickness (float): The thickness of the material in the
                construction.
        """
        self.Material = Material
        self.Thickness = Thickness

    @property
    def Material(self):
        """Get or set the material of self."""
        return self._material

    @Material.setter
    def Material(self, value):
        from archetypal.template.materials import GasMaterial

        assert isinstance(value, GasMaterial), (
            f"Input value error for '{value}'. Value must be of type (GasMaterial), "
            f"not {type(value)}."
        )
        self._material = value

    @property
    def Thickness(self):
        """Get or set the material thickness [m]."""
        return self._thickness

    @Thickness.setter
    def Thickness(self, value):
        self._thickness = value
        if value < 0.003:
            log(
                "Modeling layer thinner (less) than 0.003 m (not recommended) for "
                f"MaterialLayer '{self}'",
                lg.WARNING,
            )

    @property
    def resistivity(self):
        """Get or set the resistivity of the material layer [m-K/W]."""
        return 1 / self.Material.Conductivity

    @resistivity.setter
    def resistivity(self, value):
        self.Material.Conductivity = 1 / validators.float(value, minimum=0)

    @property
    def r_value(self):
        """Get or set the the R-value of the material layer [m2-K/W].

        Note that, when setting the R-value, the thickness of the material will
        be adjusted and the conductivity will remain fixed.
        """
        return self.Thickness / self.Material.Conductivity

    @r_value.setter
    def r_value(self, value):
        self.Thickness = validators.float(value, minimum=0) * self.Material.Conductivity

    @property
    def u_value(self):
        """Get or set the heat transfer coefficient [W/(m2⋅K)]."""
        return 1 / self.r_value

    @u_value.setter
    def u_value(self, value):
        self.r_value = 1 / validators.float(value, minimum=0)

    @property
    def heat_capacity(self):
        """Get the material layer's heat capacity [J/(m2-k)]."""
        return (
            self.Material.Density
            * self.Material.specific_heat_at_temperature(273.15)
            * self.Thickness
        )

    @property
    def specific_heat(self):
        """Get the material layer's specific heat at 0C [J/kg-K]."""
        return self.Material.specific_heat

    def u_value(
        self,
        delta_t=15,
        emissivity_1=0.84,
        emissivity_2=0.84,
        height=1.0,
        t_kelvin=273.15,
        pressure=101325,
    ):
        """Get the U-value of a vertical gas cavity given temp diff and emissivity.

        Args:
            delta_t: The temperature difference across the gas cavity [C]. This
                influences how strong the convection is within the gas gap. Default is
                15C, which is consistent with the NFRC standard for double glazed units.
            emissivity_1: The emissivity of the surface on one side of the cavity.
                Default is 0.84, which is typical of clear, uncoated glass.
            emissivity_2: The emissivity of the surface on the other side of the cavity.
                Default is 0.84, which is typical of clear, uncoated glass.
            height: An optional height for the cavity in meters. Default is 1.0,
                which is consistent with NFRC standards.
            t_kelvin: The average temperature of the gas cavity in Kelvin.
                Default: 273.15 K (0C).
            pressure: The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        return self.convective_conductance(
            delta_t, height, t_kelvin, pressure
        ) + self.radiative_conductance(emissivity_1, emissivity_2, t_kelvin)

    def u_value_at_angle(
        self,
        delta_t=15,
        emissivity_1=0.84,
        emissivity_2=0.84,
        height=1.0,
        angle=90,
        t_kelvin=273.15,
        pressure=101325,
    ):
        """Get the U-value of a vertical gas cavity given temp diff and emissivity.

        Args:
            delta_t: The temperature difference across the gas cavity [C]. This
                influences how strong the convection is within the gas gap. Default is
                15C, which is consistent with the NFRC standard for double glazed units.
            emissivity_1: The emissivity of the surface on one side of the cavity.
                Default is 0.84, which is typical of clear, uncoated glass.
            emissivity_2: The emissivity of the surface on the other side of the cavity.
                Default is 0.84, which is typical of clear, uncoated glass.
            height: An optional height for the cavity in meters. Default is 1.0,
                which is consistent with NFRC standards.
            angle: An angle in degrees between 0 and 180.
                0 = A horizontal cavity with downward heat flow through the layer.
                90 = A vertical cavity
                180 = A horizontal cavity with upward heat flow through the layer.
            t_kelvin: The average temperature of the gas cavity in Kelvin.
                Default: 273.15 K (0C).
            pressure: The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        return self.convective_conductance_at_angle(
            delta_t, height, angle, t_kelvin, pressure
        ) + self.radiative_conductance(emissivity_1, emissivity_2, t_kelvin)

    def convective_conductance(
        self, delta_t=15, height=1.0, t_kelvin=273.15, pressure=101325
    ):
        """Get convective conductance of the cavity in a vertical position.

        Args:
            delta_t: The temperature difference across the gas cavity [C]. Default is
                15C, which is consistent with the NFRC standard for double glazed units.
            height: An optional height for the cavity in meters. Default is 1.0,
                which is consistent with NFRC standards.
            t_kelvin: The average temperature of the gas cavity in Kelvin.
                Default: 273.15 K (0C).
            pressure: The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        return self.nusselt(delta_t, height, t_kelvin, pressure) * (
            self.Material.conductivity_at_temperature(t_kelvin) / self.Thickness
        )

    def convective_conductance_at_angle(
        self, delta_t=15, height=1.0, angle=90, t_kelvin=273.15, pressure=101325
    ):
        """Get convective conductance of the cavity in an angle.

        Args:
            delta_t: The temperature difference across the gas cavity [C]. Default is
                15C, which is consistent with the NFRC standard for double glazed units.
            height: An optional height for the cavity in meters. Default is 1.0,
                which is consistent with NFRC standards.
            angle: An angle in degrees between 0 and 180.
                * 0 = A horizontal cavity with downward heat flow through the layer.
                * 90 = A vertical cavity
                * 180 = A horizontal cavity with upward heat flow through the layer.
            t_kelvin: The average temperature of the gas cavity in Kelvin.
                Default: 273.15 K (0C).
            pressure: The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        return self.nusselt_at_angle(delta_t, height, angle, t_kelvin, pressure) * (
            self.Material.conductivity_at_temperature(t_kelvin) / self.Thickness
        )

    def radiative_conductance(
        self, emissivity_1=0.84, emissivity_2=0.84, t_kelvin=273.15
    ):
        """Get the radiative conductance of the cavity given emissivities on both sides.

        Args:
            emissivity_1: The emissivity of the surface on one side of the cavity.
                Default is 0.84, which is typical of clear, uncoated glass.
            emissivity_2: The emissivity of the surface on the other side of the cavity.
                Default is 0.84, which is typical of clear, uncoated glass.
            t_kelvin: The average temperature of the gas cavity in Kelvin.
                Default: 273.15 K (0C).
        """
        return (
            (4 * 5.6697e-8)
            * (((1 / emissivity_1) + (1 / emissivity_2) - 1) ** -1)
            * (t_kelvin ** 3)
        )

    def nusselt_at_angle(
        self, delta_t=15, height=1.0, angle=90, t_kelvin=273.15, pressure=101325
    ):
        """Get Nusselt number for a cavity at a given angle, temp diff and height.

        Args:
            delta_t: The temperature difference across the gas cavity [C]. Default is
                15C, which is consistent with the NFRC standard for double glazed units.
            height: An optional height for the cavity in meters. Default is 1.0,
                which is consistent with NFRC standards.
            angle: An angle in degrees between 0 and 180.
                * 0 = A horizontal cavity with downward heat flow through the layer.
                * 90 = A vertical cavity
                * 180 = A horizontal cavity with upward heat flow through the layer.
            t_kelvin: The average temperature of the gas cavity in Kelvin.
                Default: 273.15 K (0C).
            pressure: The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """

        def dot_x(x):
            return (x + abs(x)) / 2

        rayleigh = self.rayleigh(delta_t, t_kelvin, pressure)
        if angle < 60:
            cos_a = math.cos(math.radians(angle))
            sin_a_18 = math.sin(1.8 * math.radians(angle))
            term_1 = dot_x(1 - (1708 / (rayleigh * cos_a)))
            term_2 = 1 - ((1708 * (sin_a_18 ** 1.6)) / (rayleigh * cos_a))
            term_3 = dot_x(((rayleigh * cos_a) / 5830) ** (1 / 3) - 1)
            return 1 + (1.44 * term_1 * term_2) + term_3
        elif angle < 90:
            g = 0.5 / ((1 + ((rayleigh / 3160) ** 20.6)) ** 0.1)
            n_u1 = (1 + (((0.0936 * (rayleigh ** 0.314)) / (1 + g)) ** 7)) ** (1 / 7)
            n_u2 = (0.104 + (0.175 / (self.Thickness / height))) * (rayleigh ** 0.283)
            n_u_60 = max(n_u1, n_u2)
            n_u_90 = self.nusselt(delta_t, height, t_kelvin, pressure)
            return (n_u_60 + n_u_90) / 2
        elif angle == 90:
            return self.nusselt(delta_t, height, t_kelvin, pressure)
        else:
            n_u_90 = self.nusselt(delta_t, height, t_kelvin, pressure)
            return 1 + ((n_u_90 - 1) * math.sin(math.radians(angle)))

    def nusselt(self, delta_t=15, height=1.0, t_kelvin=273.15, pressure=101325):
        """Get Nusselt number for a vertical cavity given the temp diff. and height.

        Args:
            delta_t: The temperature difference across the gas cavity [C]. Default is
                15C, which is consistent with the NFRC standard for double glazed units.
            height: An optional height for the cavity in meters. Default is 1.0,
                which is consistent with NFRC standards.
            t_kelvin: The average temperature of the gas cavity in Kelvin.
                Default: 273.15 K (0C).
            pressure: The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        rayleigh = self.rayleigh(delta_t, t_kelvin, pressure)
        if rayleigh > 50000:
            n_u_l_1 = 0.0673838 * (rayleigh ** (1 / 3))
        elif rayleigh > 10000:
            n_u_l_1 = 0.028154 * (rayleigh ** 0.4134)
        else:
            n_u_l_1 = 1 + 1.7596678e-10 * (rayleigh ** 2.2984755)
        n_u_l_2 = 0.242 * ((rayleigh * (self.Thickness / height)) ** 0.272)
        return max(n_u_l_1, n_u_l_2)

    def rayleigh(self, delta_t=15, t_kelvin=273.15, pressure=101325):
        """Get Rayleigh number given the temperature difference across the cavity.

        Args:
            delta_t: The temperature difference across the gas cavity [C]. Default is
                15C, which is consistent with the NFRC standard for double glazed units.
            t_kelvin: The average temperature of the gas cavity in Kelvin.
                Default: 273.15 K (0C).
            pressure: The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        _numerator = (
            (self.Material.density_at_temperature(t_kelvin, pressure) ** 2)
            * (self.Thickness ** 3)
            * 9.81
            * self.Material.specific_heat_at_temperature(t_kelvin)
            * delta_t
        )
        _denominator = (
            t_kelvin
            * self.Material.viscosity_at_temperature(t_kelvin)
            * self.Material.conductivity_at_temperature(t_kelvin)
        )
        return _numerator / _denominator

    def to_dict(self):
        """Return MaterialLayer dictionary representation."""
        return collections.OrderedDict(
            Material={"$ref": str(self.Material.id)},
            Thickness=round(self.Thickness, decimals=3),
        )

    def to_epbunch(self, idf):
        """Convert self to an EpBunch given an IDF model.

        Notes:
            The object is added to the idf model.
            The thickness is passed to the epbunch.

        Args:
            idf (IDF): An IDF model.

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        return self.Material.to_epbunch(idf, self.Thickness)

    def mapping(self):
        """Get a dict based on the object properties, useful for dict repr."""
        return dict(Material=self.Material, Thickness=self.Thickness)

    def get_unique(self):
        """Return the first of all the created objects that is equivalent to self."""
        return self

    def __hash__(self):
        """Return the hash value of self."""
        return hash(id(self))

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, GasLayer):
            return NotImplemented
        else:
            return all(
                [self.Thickness == other.Thickness, self.Material == other.Material]
            )

    def __repr__(self):
        """Return a representation of self."""
        return "{} with thickness of {:,.3f} m".format(self.Material, self.Thickness)

    def __iter__(self):
        """Iterate over attributes. Yields tuple of (keys, value)."""
        for k, v in self.mapping().items():
            yield k, v

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(self.Material, self.Thickness)

    @property
    def children(self):
        return (self.Material,)



"""Slab module"""

import logging as lg
import shutil
import subprocess
import time
from threading import Thread

from packaging.version import Version
from path import Path
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from archetypal.eplus_interface.exceptions import EnergyPlusProcessError
from archetypal.eplus_interface.version import EnergyPlusVersion
from archetypal.utils import log


class SlabThread(Thread):
    """Slab program manager.

    The slab program used to calculate the results is included with the
    EnergyPlus distribution. It requires an input file named GHTin.idf in
    input data file format. The needed corresponding idd file is
    SlabGHT.idd. An EnergyPlus weather file for the location is also needed.
    """

    def __init__(self, idf, tmp):
        """Constructor."""
        super(SlabThread, self).__init__()
        self.p = None
        self.std_out = None
        self.std_err = None
        self.idf = idf
        self.cancelled = False
        self.run_dir = Path(tmp).expand()
        self.exception = None
        self.name = "RunSlab_" + self.idf.name
        self.include = None

    @property
    def cmd(self):
        """Get the command."""
        cmd_path = Path(shutil.which("Slab", path=self.run_dir))
        return [cmd_path.relpath(self.run_dir)]

    def run(self):
        """Wrapper around the EnergyPlus command line interface."""
        self.cancelled = False
        # get version from IDF object or by parsing the IDF file for it

        # Move files into place
        self.epw = self.idf.epw.copy(self.run_dir / "in.epw").expand()
        self.idfname = Path(self.idf.savecopy(self.run_dir / "in.idf")).expand()
        self.idd = self.idf.iddname.copy(self.run_dir).expand()

        # Get executable using shutil.which (determines the extension based on
        # the platform, eg: .exe. And copy the executable to tmp
        slab_exe = shutil.which("Slab", path=self.eplus_home)
        if slab_exe is None:
            log(
                f"The Slab program could not be found at " f"'{self.eplus_home}'",
                lg.WARNING,
            )
            return
        self.slabexe = Path(slab_exe).copy(self.run_dir)
        self.slabidd = (self.eplus_home / "SlabGHT.idd").copy(self.run_dir)

        # The GHTin.idf file is copied from the self.include list (added by
        # ExpandObjects. If self.include is empty, no need to run Slab.
        self.include = [Path(file).copy(self.run_dir) for file in self.idf.include]
        if not self.include:
            self.cleanup_callback()
            return

        # Run Slab Program
        with logging_redirect_tqdm(loggers=[lg.getLogger(self.idf.name)]):
            with tqdm(
                unit_scale=True,
                miniters=1,
                desc=f"RunSlab #{self.idf.position}-{self.idf.name}",
                position=self.idf.position,
            ) as progress:

                self.p = subprocess.Popen(
                    self.cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=True,  # can use shell
                    cwd=self.run_dir.abspath(),
                )
                start_time = time.time()
                self.msg_callback("Begin Slab Temperature Calculation processing . . .")
                for line in self.p.stdout:
                    self.msg_callback(line.decode("utf-8").strip("\n"))
                    progress.update()

                # We explicitly close stdout
                self.p.stdout.close()

                # Wait for process to complete
                self.p.wait()

                # Communicate callbacks
                if self.cancelled:
                    self.msg_callback("RunSlab cancelled")
                    # self.cancelled_callback(self.std_out, self.std_err)
                else:
                    if self.p.returncode == 0:
                        self.msg_callback(
                            "RunSlab completed in {:,.2f} seconds".format(
                                time.time() - start_time
                            )
                        )
                        self.success_callback()
                        for line in self.p.stderr:
                            self.msg_callback(line.decode("utf-8"))
                    else:
                        self.msg_callback("RunSlab failed")
                        self.failure_callback()

    def msg_callback(self, *args, **kwargs):
        """Pass message to logger."""
        log(*args, name=self.idf.name, **kwargs)

    def success_callback(self):
        """Parse surface temperature and append to IDF file."""
        temp_schedule = self.run_dir / "SLABSurfaceTemps.txt"
        if temp_schedule.exists():
            with open(self.idf.idfname, "a") as outfile:
                with open(temp_schedule) as infile:
                    next(infile)  # Skipping first line
                    next(infile)  # Skipping second line
                    for line in infile:
                        outfile.write(line)
            # invalidate attributes dependant on idfname, since it has changed
            self.idf._reset_dependant_vars("idfname")
        self.cleanup_callback()

    def cleanup_callback(self):
        """clean up temp files, directories and variables that need cleanup."""

        # Remove from include
        ghtin = self.idf.output_directory / "GHTIn.idf"
        if ghtin.exists():
            try:
                self.idf.include.remove(ghtin)
                ghtin.remove()
            except ValueError:
                log("nothing to remove", lg.DEBUG)

    def failure_callback(self):
        """Parse error file and log"""
        error_filename = self.run_dir / "eplusout.err"
        if error_filename.exists():
            with open(error_filename, "r") as stderr:
                stderr_r = stderr.read()
                self.exception = EnergyPlusProcessError(
                    cmd=self.cmd, stderr=stderr_r, idf=self.idf
                )
        self.cleanup_callback()

    def cancelled_callback(self, stdin, stdout):
        """Call on cancelled."""
        self.cleanup_callback()

    @property
    def eplus_home(self):
        """Get the version-dependant directory where executables are installed."""
        if self.idf.file_version <= Version("7.2"):
            install_dir = self.idf.file_version.current_install_dir / "bin"
        else:
            install_dir = (
                self.idf.file_version.current_install_dir
                / "PreProcess"
                / "GrndTempCalc"
            )
        return install_dir.expand()

    def stop(self):
        if self.p.poll() is None:
            self.msg_callback("Attempting to cancel simulation ...")
            self.cancelled = True
            self.p.kill()



"""IDF model package. Working with energy plus files"""

__all__ = [
    "__eq__",
    "_parse_idd_type",
    "get_default",
    "makedict",
    "nameexists",
    "hash_model",
    "IDF",
    "Outputs",
    "Meters",
    "Variables",
]

from .extensions import __eq__, _parse_idd_type, get_default, makedict, nameexists
from .idf import IDF
from .meters import Meters
from .outputs import Outputs
from .reports import get_ideal_loads_summary
from .util import hash_model
from .variables import Variables



"""EnergyPlusVersion module handles finding idd paths and valid versions."""

import platform
import re
import warnings
from itertools import chain

from packaging.version import Version
from path import Path

from archetypal import settings
from archetypal.eplus_interface.exceptions import (
    EnergyPlusVersionError,
    InvalidEnergyPlusVersion,
)


class EnergyPlusVersion(Version):
    """EnergyPlusVersion class.

    This class subclasses the :class:`packaging.version.Version` class. It is usuful
    to compare version numbers together.

    Any EnergyPlusVersion numbers are checked against valid versions before they can
    be initialized.

    Examples:
        To create a version number:

        >>> from archetypal.eplus_interface.version import EnergyPlusVersion
        >>> EnergyPlusVersion("9.2.0")
        <EnergyPlusVersion('9.2.0')>

        An invalid version number raises an exception:

        >>> from archetypal.eplus_interface.version import EnergyPlusVersion
        >>> EnergyPlusVersion("3.2.0")  # "3.2.0" was never released.
        archetypal.eplus_interface.exceptions.InvalidEnergyPlusVersion

    """

    __slots__ = ("_valid_paths", "_install_locations")

    def __init__(self, version):
        """Initialize an EnergyPlusVersion from a version number.

        Args:
            version (str, EnergyPlusVersion): The version number to create. Can be a
                string a tuple or another EnergyPlusVersion object.

        Raises:
            InvalidEnergyPlusVersion: If the version is not a valid version number.
        """
        self.install_locations = {}
        self.valid_idd_paths = {}

        if isinstance(version, tuple):
            version = ".".join(map(str, version[0:3]))
        if isinstance(version, Version):
            version = ".".join(map(str, (version.major, version.minor, version.micro)))
        if isinstance(version, str) and "-" in version:
            version = version.replace("-", ".")
        super(EnergyPlusVersion, self).__init__(version)
        if self.dash not in self.valid_versions:
            raise InvalidEnergyPlusVersion

    @classmethod
    def latest(cls):
        """Initialize an EnergyPlusVersion with the latest version installed."""
        eplus_homes = get_eplus_basedirs()

        # check if any EnergyPlus install exists
        if not eplus_homes:
            raise Exception(
                "No EnergyPlus installation found. Make sure you have EnergyPlus "
                "installed. Go to https://energyplus.net/downloads to download the "
                "latest version of EnergyPlus."
            )

        # Find the most recent version of EnergyPlus installed from the version
        # number (at the end of the folder name)
        version = next(
            iter(
                sorted(
                    (
                        Version(
                            re.search(r"\d+(-\d+)+", home.stem)
                            .group()
                            .replace("-", ".")
                        )
                        for home in eplus_homes
                    ),
                    reverse=True,
                )
            )
        )
        return cls(version)

    @property
    def dash(self) -> str:
        """Return the version number as a dash-separated string: "major-minor-micro"."""
        return "-".join(map(str, (self.major, self.minor, self.micro)))

    @property
    def dot(self) -> str:
        """Return the version number as a dot-separated string: "major.minor.micro"."""
        return ".".join(map(str, (self.major, self.minor, self.micro)))

    @property
    def current_idd_path(self):
        """Get the current Idd file path for this version."""
        return self.valid_idd_paths[self.dash]

    @property
    def current_install_dir(self):
        """Get the current installation directory for this EnergyPlus version."""
        try:
            return self.install_locations[self.dash]
        except KeyError:
            raise EnergyPlusVersionError(
                f"EnergyPlusVersion {self.dash} is not installed."
            )

    @property
    def tuple(self) -> tuple:
        """Return the version number as a tuple: (major, minor, micro)."""
        return self.major, self.minor, self.micro

    @property
    def valid_versions(self) -> set:
        """List the idd file version found on this machine."""
        if not self.valid_idd_paths:
            # Little hack in case E+ is not installed
            _choices = {
                settings.ep_version,
            }
        else:
            _choices = set(self.valid_idd_paths.keys())

        return _choices

    @property
    def install_locations(self) -> dict:
        """Get or set the available EnergyPlus root folders keyed by version number.

        Installation folders are detected automatically at the default location for
        all platforms.
        """
        return self._install_locations

    @install_locations.setter
    def install_locations(self, value):
        if not value:
            value = {}
            for basedir in get_eplus_basedirs():
                # match the Idd file contained in basedir
                match = re.search(r"\d+(-\d+)+", basedir)
                version = match.group()
                value[version] = basedir.expand()
        self._install_locations = value

    @property
    def valid_idd_paths(self) -> dict:
        """Get or set the idd paths as a dict with version numbers as keys."""
        return self._valid_paths

    @valid_idd_paths.setter
    def valid_idd_paths(self, value):
        assert isinstance(value, dict)
        if not value:
            try:
                basedirs_ = []
                for version, basedir in self.install_locations.items():
                    updater_ = basedir / "PreProcess" / "IDFVersionUpdater"
                    if updater_.exists():
                        basedirs_.append(updater_.files("*.idd"))
                    else:
                        # The IDFVersionUpdate folder could be removed in some
                        # installation (eg Docker container).
                        # Add the idd contained in the basedir instead.
                        basedirs_.append(basedir.files("*.idd"))
                iddnames = set(chain.from_iterable(basedirs_))
            except FileNotFoundError:
                _valid_paths = {}
            else:
                _valid_paths = {}
                for iddname in iddnames:
                    match = re.search("\d+(-\d+)+", iddname.stem)
                    if match is None:
                        # match the Idd file contained in basedir
                        match = re.search(r"\d+(-\d+)+", iddname.stem)
                        version = match.group()
                    else:
                        version = match.group()

                    _valid_paths[version] = iddname
        self._valid_paths = dict(sorted(_valid_paths.items()))

    @classmethod
    def current(cls):
        """Initialize an EnergyPlusVersion object for the specified module version.

        Notes:
            Specified by :ref:`archetypal.settings.ep_version` which looks for the
            `"ENERGYPLUS_VERSION` environment variable.
        """
        version = settings.ep_version
        return cls(version)

    @property
    def current_install_location(self):
        return self.install_locations[self.dash]

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def __copy__(self):
        """Return a copy of self."""
        return EnergyPlusVersion(self)

    def __repr__(self) -> str:
        """Return a representation of self."""
        return f"<EnergyPlusVersion('{str(self)}')>"


def get_eplus_basedirs():
    """Return a list of possible E+ install paths."""
    if platform.system() == "Windows":
        eplus_homes = Path("C:\\").dirs("EnergyPlus*")
        return eplus_homes
    elif platform.system() == "Linux":
        eplus_homes = Path("/usr/local/").dirs("EnergyPlus*")
        return eplus_homes
    elif platform.system() == "Darwin":
        eplus_homes = Path("/Applications").dirs("EnergyPlus*")
        return eplus_homes
    else:
        warnings.warn(
            "Archetypal is not compatible with %s. It is only compatible "
            "with Windows, Linux or MacOs" % platform.system()
        )


def warn_if_not_compatible():
    """Check if an EnergyPlus install is detected.

    Warnings:
        If the latest version detected is higher than the one specified by
        archetypal, a warning is raised.
    """
    eplus_homes = get_eplus_basedirs()

    if not eplus_homes:
        warnings.warn(
            "No installation of EnergyPlus could be detected on this "
            "machine. Please install EnergyPlus from https://energyplus.net before "
            "using archetypal"
        )



"""Eppy extensions module."""

import copy

from eppy.bunch_subclass import BadEPFieldError
from eppy.EPlusInterfaceFunctions.eplusdata import Eplusdata, Idd, removecomment
from eppy.idf_msequence import Idf_MSequence
from geomeppy.patches import EpBunch

from archetypal.utils import extend_class, log


@extend_class(EpBunch)
def __eq__(self: EpBunch, other):
    """Test the equality of two EpBunch objects using all attribute values."""
    if not isinstance(other, EpBunch):
        return False
    return all(str(a).upper() == str(b).upper() for a, b in zip(self.obj, other.obj))


@extend_class(EpBunch)
def nameexists(self: EpBunch):
    """Return True if EpBunch Name already exists in idf.idfobjects[KEY]."""
    existing_objs = self.theidf.idfobjects[self.key.upper()]
    try:
        return self.Name.upper() in [obj.Name.upper() for obj in existing_objs]
    except BadEPFieldError:
        return False


@extend_class(EpBunch)
def get_default(self: EpBunch, name):
    """Return the default value of a field"""
    if "default" in self.getfieldidd(name).keys():
        _type = _parse_idd_type(self, name)
        default_ = next(iter(self.getfieldidd_item(name, "default")), None)
        return _type(default_)
    else:
        return ""


@extend_class(EpBunch)
def to_dict(self: EpBunch):
    """Get the dict representation of the EpBunch."""
    return {k: v for k, v in zip(self.fieldnames, self.fieldvalues)}


@extend_class(EpBunch)
def __init__(self, obj, objls, objidd, *args, **kwargs):
    """Extension of EpBunch to add daylighting:referencepoints coords."""
    super(EpBunch, self).__init__(obj, objls, objidd, *args, **kwargs)
    if self.key.upper() == "DAYLIGHTING:REFERENCEPOINT":
        func_dict = {
            "coords": get_coords,
        }
        self.__functions.update(func_dict)


def get_coords(obj: EpBunch):
    """Return tuple of X, Y, Z"""
    return [
        (
            obj.XCoordinate_of_Reference_Point,
            obj.YCoordinate_of_Reference_Point,
            obj.ZCoordinate_of_Reference_Point,
        ),
    ]


@extend_class(Idf_MSequence)
def to_dict(self: Idf_MSequence):
    """Get the list of dict representation of the Idf_Msequence."""
    return [obj.to_dict() for obj in self]


@extend_class(Eplusdata)
def makedict(self: Eplusdata, dictfile, fnamefobject):
    """stuff file data into the blank dictionary."""
    # fname = './exapmlefiles/5ZoneDD.idf'
    # fname = './1ZoneUncontrolled.idf'
    if isinstance(dictfile, Idd):
        localidd = copy.deepcopy(dictfile)
        dt, dtls = localidd.dt, localidd.dtls
    else:
        dt, dtls = self.initdict(dictfile)
    fnamefobject.seek(0)  # make sure to read from the beginning
    astr = fnamefobject.read()
    try:
        astr = astr.decode("ISO-8859-2")
    except AttributeError:
        pass
    nocom = removecomment(astr, "!")
    idfst = nocom
    # alist = string.split(idfst, ';')
    alist = idfst.split(";")
    lss = []
    for element in alist:
        # lst = string.split(element, ',')
        lst = element.split(",")
        lss.append(lst)

    for i in range(0, len(lss)):
        for j in range(0, len(lss[i])):
            lss[i][j] = lss[i][j].strip()

    for element in lss:
        node = element[0].upper()
        if node in dt:
            # stuff data in this key
            dt[node.upper()].append(element)
        else:
            # scream
            if node == "":
                continue
            log("this node -%s-is not present in base dictionary" % node)

    self.dt, self.dtls = dt, dtls
    return dt, dtls


def _parse_idd_type(epbunch, name):
    """Parse the fieldvalue type into a python type.

    Possible types are:
        - integer -> int
        - real -> float
        - alpha -> str          (arbitrary string),
        - choice -> str         (alpha with specific list of choices, see \key)
        - object-list -> str    (link to a list of objects defined elsewhere, see \object-list and \reference)
        - external-list -> str  (uses a special list from an external source, see \external-list)
        - node -> str           (name used in connecting HVAC components)
    """
    _type = next(iter(epbunch.getfieldidd_item(name, "type")), "").lower()
    if _type == "real":
        return float
    elif _type == "alpha":
        return str
    elif _type == "integer":
        return int
    else:
        return str


# relationship between epbunch output frequency and db.
bunch2db = {
    "Detailed": ["HVAC System Timestep", "Zone Timestep"],
    "Timestep": ["HVAC System Timestep", "Zone Timestep"],
    "Hourly": "Hourly",
    "Daily": "Daily",
    "Monthly": "Monthly",
    "RunPeriod": "Run Period",
    "Environment": "Run Period",
    "Annual": "Annual",
}


def get_name_attribute(__o: EpBunch):
    try:
        return getattr(__o, "Key_Name")
    except BadEPFieldError:  # Backwards compatibility
        return getattr(__o, "Name")



"""EnergyPlus python interface."""

__all__ = [
    "InvalidEnergyPlusVersion",
    "EnergyPlusProcessError",
    "EnergyPlusVersion",
    "EnergyPlusProgram",
    "EnergyPlusVersionError",
    "EnergyPlusWeatherError",
    "BasementThread",
    "EnergyPlusExe",
    "EnergyPlusThread",
    "ExpandObjectsThread",
    "SlabThread",
    "TransitionThread",
]

from archetypal.eplus_interface.basement import BasementThread
from archetypal.eplus_interface.energy_plus import (
    EnergyPlusExe,
    EnergyPlusProgram,
    EnergyPlusThread,
)
from archetypal.eplus_interface.exceptions import (
    EnergyPlusProcessError,
    EnergyPlusVersionError,
    EnergyPlusWeatherError,
    InvalidEnergyPlusVersion,
)
from archetypal.eplus_interface.expand_objects import ExpandObjectsThread
from archetypal.eplus_interface.slab import SlabThread
from archetypal.eplus_interface.transition import TransitionThread
from archetypal.eplus_interface.version import EnergyPlusVersion
